RANK_ORDER = [
    "0", "root", "top", "superkingdom", "kingdom", "subkingdom", "superclade", "clade", "subclade",
    "superphylum", "phylum", "subphylum", "superclass", "class", "subclass", "superorder", "order",
    "suborder", "superfamily", "family", "subfamily", "supergenus", "genus", "subgenus", "species group",
    "species subgroup", "superspecies", "species", "subspecies", "serogroup", "serotype", "superstrain",
    "strain", "substrain", "no rank"
]
TAX_RESULTS_HEADER = [
    "taxid", "%ClassifiedCumulative", "ClassifiedCumulative", "ClassifiedCount",
    "%SplitCumulative", "SplitCumulative", "SplitCount", "Rank", "Lineage"
]

PHYLA_COLOURS = [
    ("Proteobacteria", "#E85343"),
    ("Bacteroidetes", "#30B674"),
    ("Firmicutes", "#6585C3"),
    ("Actinobacteria", "#FEDB45")
]
